#!/usr/bin/env python

import rospy
from radar_sensor_plugin.msg import radar_data
from std_msgs.msg import String
from nav_msgs.msg import OccupancyGrid
from nav_msgs.msg import MapMetaData
from geometry_msgs.msg import Pose
from geometry_msgs.msg import Point
from geometry_msgs.msg import Quaternion
from math import sin, cos, pi, sqrt, pow
import numpy as np


pub = rospy.Publisher('/radar_map', OccupancyGrid, queue_size = 10)     # Topic to publish occupancy map to
map_msg = OccupancyGrid()                                               # The global occupancy map published to be read by rviz
radar_max_range = 10                                                    # Max range that the radar can see
grid_resolution = 0.2                                                   # Set the grid resolution [m/cell]
grid_width = 100 / grid_resolution                                      # Set the width of the occupancy grid
radar_azimuth = 65/180.0*pi                                             # Azimuth of the radar, set to radians
detection_theshold = 0.038                                              # The amplitude for detection of objects for the radar sensor
grid_drawing_iteration = radar_max_range/grid_resolution                # Used to tweak resolution of drawing on the map, follows maps resolution

# Update the Occupancy map and publish
def occupancy_map_publisher_callback1(data):
    global map_msg, grid_width, grid_resolution, radar_azimuth

    quadrotor_position = data.position
    yaw = quadrotor_position[3] 

    # Get range at which the radar detects an object
    range_to_object = range_detection(data.histogram_bar, data.range)
    if range_to_object == None:
        range_to_object = 10

    A, B = get_outer_points(quadrotor_position, yaw, range_to_object, radar_azimuth)
    occupancy_grid_mapping(A, B, quadrotor_position, map_msg.data, range_to_object)

    pub.publish(map_msg)

def occupancy_map_publisher_callback2(data):
    global map_msg, grid_width, grid_resolution, radar_azimuth

    quadrotor_position = data.position
    yaw = quadrotor_position[3] - 1.57

    # Get range at which the radar detects an object
    range_to_object = range_detection(data.histogram_bar, data.range)
    if range_to_object == None:
        range_to_object = 10

    A, B = get_outer_points(quadrotor_position, yaw, range_to_object, radar_azimuth)
    occupancy_grid_mapping(A, B, quadrotor_position, map_msg.data, range_to_object)

def occupancy_map_publisher_callback3(data):
    global map_msg, grid_width, grid_resolution, radar_azimuth

    quadrotor_position = data.position
    yaw = quadrotor_position[3] - 1.57*2

    # Get range at which the radar detects an object
    range_to_object = range_detection(data.histogram_bar, data.range)
    if range_to_object == None:
        range_to_object = 10

    A, B = get_outer_points(quadrotor_position, yaw, range_to_object, radar_azimuth)
    occupancy_grid_mapping(A, B, quadrotor_position, map_msg.data, range_to_object)

def occupancy_map_publisher_callback4(data):
    global map_msg, grid_width, grid_resolution, radar_azimuth

    quadrotor_position = data.position
    yaw = quadrotor_position[3] + 1.57

    # Get range at which the radar detects an object
    range_to_object = range_detection(data.histogram_bar, data.range)
    if range_to_object == None:
        range_to_object = 10

    A, B = get_outer_points(quadrotor_position, yaw, range_to_object, radar_azimuth)
    occupancy_grid_mapping(A, B, quadrotor_position, map_msg.data, range_to_object)


# Listen for radar data, then calls occupancy_map_publisher_callback to update the occupancy map
def radar_mapper():
    global map_msg, grid_width, grid_resolution

    # Set up the map initial values
    map_meta_data = MapMetaData()

    # Resolution [m/cell]
    map_meta_data.resolution = grid_resolution
    map_meta_data.width = grid_width
    map_meta_data.height = grid_width

    # Set the position of the point of start for the mapping
    pose = Pose()
    point = Point()
    quat = Quaternion()
    # Have to set the start of the point in the corner of the map
    point.x = -grid_width * grid_resolution / 2              # Reverse to find the meters width the map is, then divide by 2 to find the axis intersections
    point.y = -grid_width * grid_resolution / 2
    point.z = 0

    quat.x = 0
    quat.y = 0
    quat.z = 0
    quat.w = 0

    pose.position = point
    pose.orientation = quat

    map_meta_data.origin = pose

    map_msg.info = map_meta_data

    # Create an empty map, setting all array points to 100 sets the map to being black
    map_msg.data = [51]*int(grid_width*grid_width)

    # Start the ROS node for mapping
    rospy.init_node('radar_mapping', anonymous=True)
    rospy.Subscriber('/radar1', radar_data, occupancy_map_publisher_callback1)
    rospy.Subscriber('/radar2', radar_data, occupancy_map_publisher_callback2)
    rospy.Subscriber('/radar3', radar_data, occupancy_map_publisher_callback3)
    rospy.Subscriber('/radar4', radar_data, occupancy_map_publisher_callback4)

    rospy.spin()


# Convert from a position in meters to a place on the grid
def pos_to_grid(pos_x, pos_y):
    global grid_width, grid_resolution
    max_size_area = grid_width * grid_resolution

    x_array_point = int(round((grid_width / 2.0) + (pos_x/max_size_area) * grid_width))
    y_array_point = int(round((grid_width / 2.0) + (pos_y/max_size_area) * grid_width))

    return x_array_point, y_array_point

# Return range of first hit threshold
def range_detection(histogram, range_value):
    global detection_theshold
    size_of_lists = len(histogram)
    for iterator in range(size_of_lists-1):
        if histogram[iterator] > detection_theshold and histogram[iterator+1] < histogram[iterator]:
            return range_value[iterator]


# Gets the outer poitns to the object given information about the radar sensor
def get_outer_points(position, yaw, range_object, azimuth):
    angle_a = yaw - azimuth/2.0
    angle_b = yaw + azimuth/2.0

    x_a = cos(angle_a)*range_object
    y_a = sin(angle_a)*range_object
    x_b = cos(angle_b)*range_object
    y_b = sin(angle_b)*range_object

    return [position[0] + x_a, position[1] + y_a], [position[0] + x_b, position[1] + y_b]



def unit_vector(vector):
    return vector / np.linalg.norm(vector)

def angle_between(v1, v2):
    v1_u = unit_vector(v1)
    v2_u = unit_vector(v2)
    return np.arccos(np.clip(np.dot(v1_u, v2_u), -1.0, 1.0))

def mid_finder(A, B):
    range_x = abs(A[0] - B[0])
    range_y = abs(A[1] - B[1])
    x = min(A[0], B[0]) + range_x/2.0
    y = min(A[1], B[1]) + range_y/2.0
    return x, y

# Assuming a square grid
def grid_to_pos(grid_point):
    global grid_width, grid_resolution
    max_size_area = grid_width * grid_resolution
    pos_x = grid_resolution * grid_point - max_size_area / 2.0
    return pos_x

    
def occupancy_calculation(range_point, angle_point, object_distance):
    global radar_azimuth
    omega = radar_azimuth
    epsilon = 0.03
    dirac = range_point
    theta = angle_point
    R = object_distance
    R_min = 0.0

    # Calculate probability of empty
    if dirac >= R_min and dirac <= R - epsilon:
        E_r = 1 - pow(((dirac - R_min)/(R - epsilon - R_min)), 2)
    else:
        E_r = 0

    if theta >= -omega / 2 and theta <= omega / 2:
        E_a = 1 - pow((2*theta / omega), 2)
    else:
        E_a = 0

    probability_empty = E_r * E_a

    # Calculate probability of occupied
    if dirac >= R - epsilon and dirac <= R + epsilon:
        O_r = 1 - pow(((dirac - R)/epsilon), 2)
    else:
        O_r = 0

    O_a = E_a
    
    probability_occupied = O_r * O_a

    return probability_empty, probability_occupied


def normalize_grid(occupied_grid):
    normalize_sum = 0.0

    for iterator in range(len(occupied_grid)):
        normalize_sum = normalize_sum + occupied_grid[iterator][2]
    if normalize_sum == 0:
        return
    for iterator in range(len(occupied_grid)):
        occupied_grid[iterator][2] = occupied_grid[iterator][2] / normalize_sum


def combine_occupied_probability(occupied_grid, grid):
    global grid_width
    for iterator in range(len(occupied_grid)):
        x = occupied_grid[iterator][0]
        y = occupied_grid[iterator][1]
        grid[int(x + grid_width * y)] = grid[int(x + grid_width * y)] + occupied_grid[iterator][2]*100 - grid[int(x + grid_width * y)] * occupied_grid[iterator][2]


def occupancy_grid_mapping(A, B, position, grid, object_distance):
    global grid_width

    # Find the location of the midpoint of the beam
    mid_x_pos, mid_y_pos = mid_finder(A, B)

    # Set the vector for the extra length of the circle and the corresponding point
    length_percentage_circle = 0.18568
    vector3_pos              = [mid_x_pos - position[0], mid_y_pos - position[1]]
    vector3_pos[0]           = vector3_pos[0] * length_percentage_circle
    vector3_pos[1]           = vector3_pos[1] * length_percentage_circle
    circle_pos               = (mid_x_pos + vector3_pos[0], mid_y_pos + vector3_pos[1])

    # Find corner points of the rectangle
    left_x_pos   = min(A[0], B[0], position[0], circle_pos[0])
    bottom_y_pos = min(A[1], B[1], position[1], circle_pos[1])
    top_y_pos    = max(A[1], B[1], position[1], circle_pos[1])
    right_x_pos  = max(A[0], B[0], position[0], circle_pos[0])

    # Getting the grid positions from the position data
    lower_grid   = pos_to_grid(left_x_pos, bottom_y_pos)
    upper_grid   = pos_to_grid(right_x_pos, top_y_pos)

    # Setting the iterator range for traversing the grid
    delta_x_grid  = upper_grid[0] - lower_grid[0]
    delta_y_grid  = upper_grid[1] - lower_grid[1]

    # Initialize list for occupied values
    occupied_grid = []
    # Iterate through the grid
    for iterator_x in range(delta_x_grid):
        for iterator_y in range(delta_y_grid):
            x = iterator_x + lower_grid[0]
            y = iterator_y + lower_grid[1]

            vector1_pos = (grid_to_pos(x) - position[0], grid_to_pos(y) - position[1])
            vector2_pos = (mid_x_pos - position[0], mid_y_pos - position[1])

            range_point = sqrt(pow(vector1_pos[0], 2) + pow(vector1_pos[1], 2))
            angle_point = angle_between(vector1_pos, vector2_pos) 

            probability_empty, probability_occupied = occupancy_calculation(range_point, angle_point, object_distance)
            
            if probability_occupied != 0:
                probability_occupied = probability_occupied * (grid[int(x + grid_width * y)]/100)
                occupied_grid.append([x, y, probability_occupied]) 

            # Superposition of empty cell
            probability_empty = (1 - grid[int(x + grid_width * y)]/100.0) + probability_empty - (1 - grid[int(x + grid_width * y)]/100.0) * probability_empty
            grid[int(x + grid_width * y)] = 100 - probability_empty*100

    if len(occupied_grid) != 0:
        normalize_grid(occupied_grid)
        combine_occupied_probability(occupied_grid, grid)

# Try running the node
if __name__ == '__main__':
    try:
        radar_mapper()
    except rospy.ROSInterruptException:
        pass
