//=================================================================================================
// Copyright (c) 2012, Johannes Meyer, TU Darmstadt
// All rights reserved.

// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions are met:
//     * Redistributions of source code must retain the above copyright
//       notice, this list of conditions and the following disclaimer.
//     * Redistributions in binary form must reproduce the above copyright
//       notice, this list of conditions and the following disclaimer in the
//       documentation and/or other materials provided with the distribution.
//     * Neither the name of the Flight Systems and Automatic Control group,
//       TU Darmstadt, nor the names of its contributors may be used to
//       endorse or promote products derived from this software without
//       specific prior written permission.

// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY
// DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
// (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
// ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
// (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
// SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//=================================================================================================

#include <gazebo_ros_radar.h>
#include <gazebo/common/Events.hh>
#include <gazebo/physics/physics.hh>
#include <gazebo/sensors/RaySensor.hh>

#include <gazebo/gazebo_config.h>
#include <radar_sensor_plugin/radar_data.h>

#include <algorithm>
#include <math.h>
#include <iterator>
#include <random>
#include <map>
#include <array>

#define PI 3.14159265359                        
#define AZIMUTH_OPENING 65                              // Size of the radar sensor opening, given in degrees
#define HISTOGRAM_RESOLUTION 20                         // The resolution of publishing histogram, the higher resolution the more spikes in the graph

namespace gazebo {

GazeboRosRadar::GazeboRosRadar()
{
   initialize_sensor_iteration_ = 0;
}

////////////////////////////////////////////////////////////////////////////////
// Destructor
GazeboRosRadar::~GazeboRosRadar()
{
  sensor_->SetActive(false);

  node_handle_->shutdown();
  delete node_handle_;
}

////////////////////////////////////////////////////////////////////////////////
// Load the controller
void GazeboRosRadar::Load(sensors::SensorPtr _sensor, sdf::ElementPtr _sdf)
{
  // Get then name of the parent sensor
#if (GAZEBO_MAJOR_VERSION > 6)
  sensor_ = std::dynamic_pointer_cast<sensors::RaySensor>(_sensor);
#else
  gzthrow("GazeboRosRadar requires a version of Gazebo of 7 or higher");
  return;
#endif
  if (!sensor_)
  {
    gzthrow("GazeboRosRadar requires a Ray Sensor as its parent");
    return;
  }

  // Get the world name.
#if (GAZEBO_MAJOR_VERSION > 6)
  std::string worldName = sensor_->WorldName();
#else
  std::string worldName = sensor_->GetWorldName();
#endif

  world_ = physics::get_world(worldName);

  // default parameters
  namespace_.clear();
  topic_ = "radar";
  frame_id_ = "/radar_link";

  // load parameters
  if (_sdf->HasElement("robotNamespace"))
    namespace_ = _sdf->GetElement("robotNamespace")->GetValue()->GetAsString();

  if (_sdf->HasElement("frameId"))
    frame_id_ = _sdf->GetElement("frameId")->GetValue()->GetAsString();

  msg_.header.frame_id = frame_id_;

  // Make sure the ROS node for Gazebo has already been initialized
  if (!ros::isInitialized())
  {
    ROS_FATAL_STREAM("A ROS node for Gazebo has not been initialized, unable to load plugin. "
      << "Load the Gazebo system plugin 'libgazebo_ros_api_plugin.so' in the gazebo_ros package)");
    return;
  }

  node_handle_ = new ros::NodeHandle(namespace_);
  publisher_ = node_handle_->advertise<radar_sensor_plugin::radar_data>(topic_, 1);

  Reset();

  this->updateConnection = this->sensor_->ConnectUpdated(
      std::bind(&GazeboRosRadar::Update, this));

  // Get number of rays the sensor is using
  ray_total_count_ = sensor_->LaserShape()->GetSampleCount() * sensor_->LaserShape()->GetVerticalSampleCount();

  // activate RaySensor
  sensor_->SetActive(true);
}

void GazeboRosRadar::Reset()
{
}

////////////////////////////////////////////////////////////////////////////////
// Update the controller
void GazeboRosRadar::Update()
{
  common::Time sim_time = world_->GetSimTime();

  // activate RaySensor if it is not yet active
  if (!sensor_->IsActive()) sensor_->SetActive(true);
  
  std::vector<std::vector<double>> array_ray_info_m(ray_total_count_, std::vector<double>(2, 0));

  for (int i = 0; i < ray_total_count_; ++i) {
    array_ray_info_m[i][0] = sensor_->LaserShape()->GetRange(i);
    array_ray_info_m[i][1] = sensor_->LaserShape()->GetRetro(i);
  }

  //Create rotated ray map
  std::map<int, std::array<double, 3> > map_ray_angles_m;

  // Get the rotation of the quadrotor in its yaw angle
  double quadrotor_yaw_m = world_->ModelByName("quadrotor")->WorldPose().Rot().Yaw(); // TODO: Get name of model from SDF if possible 

  FillRayMap(map_ray_angles_m, AZIMUTH_OPENING, ray_total_count_); 
  RotateRayMap(map_ray_angles_m, quadrotor_yaw_m, ray_total_count_);

  
  //Find which objects are in view of the sensor double roll, pitch, yaw, vector_angle, signal_strength;
  std::array<double, 3>         array_object_vector_m;
  std::array<double, 3>         array_ray_vector_m;
  std::string                   collision_object_name_m;
  double                        object_yaw_m;
  double                        vector_angle_m;
  double                        signal_strength_m;

  for (int i = 0; i < ray_total_count_; ++i)
  {
    // Have to let the sensor initialize fully, takes some iterations in Gazebo
    if (initialize_sensor_iteration_ > 120) 
    {
      collision_object_name_m =  sensor_->LaserShape()->Ray(i)->CollisionName();     

      // Checks if the colliding object is a known object TODO: See if able to be more general than saving multiple objects 
      if (collision_object_name_m.find("unit_box") != std::string::npos)
      {
        const ignition::math::Pose3d kObjectPose = world_->ModelByName("unit_box")->WorldPose();
        object_yaw_m = kObjectPose.Rot().Yaw();
        // Have to rotate 90 degrees as it has been defined somewhat confusing in Gazebo
        object_yaw_m += 3.14/2;
        // Find unit vector of object
        array_object_vector_m[0] = 1;
        array_object_vector_m[1] = 0;
        array_object_vector_m[2] = 0; 

        RotateVector(array_object_vector_m, object_yaw_m);

        // Calculate angle between vector, check if angle is correct
        array_ray_vector_m = map_ray_angles_m[i];
        vector_angle_m = CalculateVectorAngle(array_object_vector_m[0],array_object_vector_m[1],array_object_vector_m[2], array_ray_vector_m[0],array_ray_vector_m[1],array_ray_vector_m[2]);

        if (vector_angle_m > 3.14/2)
        {
            array_object_vector_m[0] = -array_object_vector_m[0];
            array_object_vector_m[1] = -array_object_vector_m[1];
            array_object_vector_m[2] = -array_object_vector_m[2];        
            vector_angle_m = CalculateVectorAngle(array_object_vector_m[0],array_object_vector_m[1],array_object_vector_m[2], array_ray_vector_m[0],array_ray_vector_m[1],array_ray_vector_m[2]);
        }

        // Weigh the received signal 
        signal_strength_m = array_ray_info_m[i][1] * cos(vector_angle_m) * (array_ray_info_m[i][0]/sensor_->RangeMax()) * (array_ray_info_m[i][0]/sensor_->RangeMax());
        array_ray_info_m[i][1] = signal_strength_m;
      }
      else 
      {
        // TODO: Calculate for normal objects
      }
    }
    else
        initialize_sensor_iteration_++;
  }
  
  std::sort(std::begin(array_ray_info_m), std::end(array_ray_info_m));

  PublishRadarData(HISTOGRAM_RESOLUTION, array_ray_info_m);
}

////////////////////////////////////////////////////////////////////////////////
// Creates a ray map around the x direction, given an azimuth opening.
void GazeboRosRadar::FillRayMap(std::map<int, std::array<double, 3> >& ray_angle_map, double azimuth_opening, double ray_count)
{
  double square_size = sqrt(ray_count);
  double length    = tan((azimuth_opening/2)*(PI/180));
  double increment = 2*length/(square_size - 1);
  double z         = length;
  double y;

  for (int i = 0; i < square_size; ++i)
  {
    y = length;
    for (int j = 0; j < square_size; ++j)
    {
        ray_angle_map[i*square_size + j] = {{1, y, z}};
        y -= increment;
    }
    z -= increment;
  }
}

////////////////////////////////////////////////////////////////////////////////
// Calculates the angle between two 3D vectors. 
double GazeboRosRadar::CalculateVectorAngle(double r1, double r2, double r3, double u1, double u2, double u3)
{
  // Need to invert the ray vector:
  double q1;
  double q2;
  double theta;

  q1 = r1*u1 + r2*u2 + r3*u3; 
  q2 = sqrt(u1*u1 + u2*u2 + u3*u3) * sqrt(r1*r1 + r2*r2 + r3*r3);
  theta  = acos(q1/q2);

  return theta;
}

////////////////////////////////////////////////////////////////////////////////
// Rotates a map filles with rays which are three dimensional. Rotates around the third dimension (yaw). Input is in radians. Square_size is the size of one of the sides.
void GazeboRosRadar::RotateRayMap(std::map<int, std::array<double, 3> >& ray_angle_map, double yaw, double ray_count)
{
  double square_size = sqrt(ray_count);
  double temp_rotated_vector[3];
  double rotation_matrix[3][3]    = {{cos(yaw), -sin(yaw), 0},
                                     {sin(yaw), cos(yaw), 0},
                                     {0, 0, 1}};

  for (int i = 0; i < square_size*square_size; ++i)
  {
    for (int j = 0; j < 3; ++j)
    {   
        temp_rotated_vector[j] = rotation_matrix[j][0]*ray_angle_map[i][0] + rotation_matrix[j][1]*ray_angle_map[i][1] + rotation_matrix[j][2]*ray_angle_map[i][2];
    }
    for (int j = 0; j < 3; ++j)
    {   
        ray_angle_map[i][j] = temp_rotated_vector[j];
    }
  }   
}

////////////////////////////////////////////////////////////////////////////////
// Rotates a 3 dimensional vector by a given angle around its third vector (yaw). Input is in radians.
void GazeboRosRadar::RotateVector(std::array<double,3>& vector, double yaw)
{
  double temp_rotated_vector[3];
  double rotation_matrix[3][3] = {{cos(yaw), -sin(yaw), 0},
                                  {sin(yaw), cos(yaw), 0},
                                  {0, 0, 1}};
  for (int j = 0; j < 3; ++j)
  {
    temp_rotated_vector[j] = rotation_matrix[j][0]*vector[0] + rotation_matrix[j][1]*vector[1] + rotation_matrix[j][2]*vector[2];
  }     
  vector[0] = temp_rotated_vector[0];
  vector[1] = temp_rotated_vector[1]; 
  vector[2] = temp_rotated_vector[2]; 
}

////////////////////////////////////////////////////////////////////////////////
// Publishes radar data given a filled in ray map, adding noise the the measurements
// The data given as parameter is already sorted
void GazeboRosRadar::PublishRadarData(double resolution, std::vector<std::vector<double>> array_ray_info)
{
  int    histogram_bar_quantity         = sensor_->RangeMax() * resolution;
  double interval_size                  = 1/resolution;
  double current_max_interval           = 0 + interval_size;
  double radar_intensity_total          = 0;

  std::vector<double>::iterator it2;

  for(int i = 0; i < ray_total_count_; i++)
  {
    if (array_ray_info[i][0] < 0.7)
      continue;

    while(array_ray_info[i][0] > current_max_interval)
    {
      current_max_interval += interval_size;
      msg_.histogram_bar.push_back(radar_intensity_total/current_max_interval);
      radar_intensity_total = 0;
      continue;
    }

    // Remove unrealistic and pre-made reflection coefficients, sum up laser reflection value
    // Creating the y axis, the amplitude of a range hit
    if (array_ray_info[i][1] > 0 && array_ray_info[i][1] < 100)
      radar_intensity_total += array_ray_info[i][1];
  }
  
  // Padding, fills up to match the dimensions of the arrays
  int padding_left = histogram_bar_quantity - msg_.histogram_bar.size();
  for(int i = 0; i < padding_left; ++i)
  {
    msg_.histogram_bar.push_back(0);
  }
  
  // Create the x axis, range
  for(double i = interval_size/2; i < sensor_->RangeMax(); i += interval_size)
  {
    msg_.range.push_back(i);
  }

  const double kMean = 0.0;
  const double kStdDev = 0.01;
  std::mt19937 generator(std::random_device{}()); 
  std::normal_distribution<double> dist(kMean, kStdDev);
 
  // Adding noise to the measurements 
  for(std::vector<double>::iterator it = msg_.histogram_bar.begin(); it != msg_.histogram_bar.end(); ++it)
  {
    *it += dist(generator);
  }
  
  double quadrotor_position_x = world_->ModelByName("quadrotor")->WorldPose().Pos().X();
  double quadrotor_position_y = world_->ModelByName("quadrotor")->WorldPose().Pos().Y();
  double quadrotor_position_z = world_->ModelByName("quadrotor")->WorldPose().Pos().Z();
  double quadrotor_position_yaw = world_->ModelByName("quadrotor")->WorldPose().Rot().Yaw();  
  double quadrotor_position_qw = world_->ModelByName("quadrotor")->WorldPose().Rot().W();
  double quadrotor_position_qx = world_->ModelByName("quadrotor")->WorldPose().Rot().X();
  double quadrotor_position_qy = world_->ModelByName("quadrotor")->WorldPose().Rot().Y();
  double quadrotor_position_qz = world_->ModelByName("quadrotor")->WorldPose().Rot().Z();

  msg_.position.push_back(quadrotor_position_x);
  msg_.position.push_back(quadrotor_position_y);
  msg_.position.push_back(quadrotor_position_z);
  msg_.position.push_back(quadrotor_position_yaw);
  msg_.position.push_back(quadrotor_position_qw);
  msg_.position.push_back(quadrotor_position_qx);
  msg_.position.push_back(quadrotor_position_qy);
  msg_.position.push_back(quadrotor_position_qz);


  msg_.header.stamp.sec  = (world_->GetSimTime()).sec;
  msg_.header.stamp.nsec = (world_->GetSimTime()).nsec;

  publisher_.publish(msg_);
  msg_.range.clear();
  msg_.histogram_bar.clear();    
  msg_.position.clear();    
}

// Register this plugin with the simulator
GZ_REGISTER_SENSOR_PLUGIN(GazeboRosRadar)

} // namespace gazebo
